#include <avr/io.h>
#include <util/delay.h>

#define EN1 PD0
#define EN2 PD1
#define EN3 PD2
#define EN4 PD3
#define SDI PD4
#define SCK PD5
#define SYNC PD7

#define CS1 PB0
#define CS2 PB1

void writeDAC(uint8_t chip, uint16_t val){
    PORTD &= ~(1 << SCK);
    PORTB &= ~(1 << chip);

    // Config bits
    const uint8_t config = 0b0011;
    for(int i = 3; i >= 0; i--){
        PORTD &= ~(1 << SDI);

        if(config & (1 << i)){
            PORTD |= (1 << SDI);
        }

        PORTD |= (1 << SCK);
        _delay_us(1);
        PORTD &= ~(1 << SCK);
    }

    // Data value
    for(int i = 11; i >= 0; i--){
        PORTD &= ~(1 << SDI);

        if(val & (1 << i)){
            PORTD |= (1 << SDI);
        }

        PORTD |= (1 << SCK);
        _delay_us(1);
        PORTD &= ~(1 << SCK);
    }

    PORTB |= (1 << chip);
}

void syncDAC(){
    PORTD &= ~(1 << SYNC);
    _delay_us(1000);
    PORTD |= ~(1 << SYNC);
}

#define MAX 4
//                                100%   92.39% 70.71% 38.27% 0%
const uint16_t currentLimits[] = {0xFFF, 0xBFF, 0x9FF, 0x8AA, 0x000};
void quarterStep(uint16_t steps){
    int8_t coil1 = 0, coil2 = MAX;
    int8_t dir1 = 1, dir2 = -1;

    for(int i = 0; i < steps; i++){
        // Guard Rails
        if(coil1 >= MAX){
            coil1 = MAX;
            dir1 = -1;
        } else if (coil1 <= MAX * -1){
            coil1 = MAX * -1;
            dir1 = 1;
        }

        if(coil2 >= MAX){
            coil2 = MAX;
            dir2 = -1;
        } else if (coil2 <= MAX * -1){
            coil2 = MAX * -1;
            dir2 = 1;
        }

        writeDAC(CS1, currentLimits[coil1 < 0 ? -coil1 : coil1]);
        writeDAC(CS2, currentLimits[coil2 < 0 ? -coil2 : coil2]);

        // Disable coil actions
        PORTD |= (1 << EN1) | (1 << EN2) | (1 << EN3) | (1 << EN4);
        syncDAC();

        // Activate in right direction
        if(coil1 < 0){
            PORTD &= ~(1 << EN2);
        } else {
            PORTD &= ~(1 << EN1);
        }

        if(coil2 < 0){
            PORTD &= ~(1 << EN4);
        } else {
            PORTD &= ~(1 << EN3);
        }

        coil1 += dir1;
        coil2 += dir2;

        _delay_us(3000);
    }

    // Disable coil actions
    PORTD |= (1 << EN1) | (1 << EN2) | (1 << EN3) | (1 << EN4);
}

void fullWave(uint8_t steps, int8_t dir){
    writeDAC(CS1, 0xFFF);
    writeDAC(CS2, 0xFFF);
    syncDAC();

    const uint8_t order[] = {EN1, EN3, EN2, EN4};

    uint8_t currentStep = 0;
    for(int i = 0; i < steps; i++){
        PORTD &= ~(1 << order[currentStep]);

        currentStep += dir;
        if(currentStep > 3){
            if(dir < 0){ // Negative underflow
                currentStep = 3;
            } else {
                currentStep = 0;
            }
        }

        _delay_us(5000);

        // Disable coil actions
        PORTD |= (1 << EN1) | (1 << EN2) | (1 << EN3) | (1 << EN4);
    }
}

void fullStep(uint8_t steps, int8_t dir){
    writeDAC(CS1, 0xFFF);
    writeDAC(CS2, 0xFFF);
    syncDAC();

    const uint8_t order[][2] = {{EN1, EN4}, {EN1, EN3}, {EN3, EN2}, {EN2, EN4}};

    uint8_t currentStep = 0;
    for(int i = 0; i < steps; i++){
        PORTD &= ~((1 << order[currentStep][0]) | (1 << order[currentStep][1]));

        currentStep += dir;
        if(currentStep > 3){
            if(dir < 0){ // Negative underflow
                currentStep = 3;
            } else {
                currentStep = 0;
            }
        }

        _delay_us(5000);

        // Disable coil actions
        PORTD |= (1 << EN1) | (1 << EN2) | (1 << EN3) | (1 << EN4);
    }
}

void halfStep(uint8_t steps, int8_t dir){
    writeDAC(CS1, 0xFFF);
    writeDAC(CS2, 0xFFF);
    syncDAC();

    const int8_t order[][2] = {{EN1, -1}, {EN1, EN3}, {EN3, -1}, {EN3, EN2}, {EN2 ,-1}, {EN2, EN4}, {EN4, -1}, {EN1, EN4}};

    uint8_t currentStep = 0;
    for(int i = 0; i < steps; i++){
        PORTD &= ~((1 << order[currentStep][0]) | (1 << order[currentStep][1]));

        currentStep += dir;
        if(currentStep > 7){
            if(dir < 0){ // Negative underflow
                currentStep = 7;
            } else {
                currentStep = 0;
            }
        }

        _delay_us(5000);

        // Disable coil actions
        PORTD |= (1 << EN1) | (1 << EN2) | (1 << EN3) | (1 << EN4);
    }
}

int main(void){
    // Set HIGH initially (enable is active LOW)
    PORTD = 0xFF;
    DDRD = 0xFF;

    PORTB = (1 << CS1) | (1 << CS2);
    DDRB = (1 << CS1) | (1 << CS2);

    writeDAC(CS1, 0xFFF);
    writeDAC(CS2, 0xFFF);
    syncDAC();

    // Buttons
    PORTB |= (1 << PB2) | (1 << PB3);
    while(1){
        if(!(PINB & (1 << PB2))){
            //quarterStep(800);
            halfStep(100, 1);
        } else if(!(PINB & (1 << PB3))){
            halfStep(100, -1);
        }

        _delay_us(10000);
    }
}
